#!/usr/bin/python
import sys
import itertools
import numpy as np
from heapq import nlargest
from math import sqrt 
import spinmatrices

'''
 Written by Andy Ohlin
 Umea University, Sweden
 
 20/11/2019
 * Can handle mixed valence
 * Systems are defined in a module file (spinmatrices.py)
 * DO NOT EDIT THIS FILE
 
 Usage Examples:
 
 general command: program isomer terse mode
 * isomer can be any system defined in spinmatrices.py, such as a which is presently corresponding to the alpha-epsilon isomers of the Keggin ion, 
 or mini, which is a test system. ALL systems should be defined in spinmatrices.py
 * terse is 0 or 1. If 1, then all duplicates (i.e. energetically degenerate states) with the same J formulae are remove
 * mode is 0 or 1. If 0, J formulae are generated. If 1, energies are calculated, with error propagation, using provided coupling constants.
 
 Examples:
 python J_generator.py a 0  0 -- compute all permutations of couplings for isomer a (alpha). 
 python J_generator.py a 1  0 -- compute all permutations of couplings, but removes all duplicates, whether incidental or symmetry-based. Terse=1
 python J_generator.py a 0  1 -- compute energies from fitted parameters. Terse=0 will allow duplicates.

'''

def GenSpecies(J):
	AllPermutationsX = list(itertools.product([1,-1],repeat=len(J))) 
	AllSpecies=[]
	for n in range(0,len(AllPermutationsX)):
		Spins=[]
		for m in range(0,len(AllPermutationsX[n])):
			Spins+=[AllPermutationsX[n][m]*SpinArray[m]]
		AllSpecies+=[Spins]
	return AllSpecies

def getMax(J):
	maximum=0
	for n in J:
		for m in n:
			if m>maximum:
				maximum=m
	return maximum

def GenCoupling(n,J):
	JList=[0]*(getMax(J)+1)

	for m in range(0,len(n)):
		for k in range(0,len(n)):
			if not (m==k) and not(n[m]==n[k]) and not (abs(n[m])/abs(n[k])==(n[m]/n[k])):
				JList[J[m][k]]= JList[J[m][k]] + 2*(SpinArray[m]*SpinArray[k])+SpinArray[k]

	for l in range(0,len(JList)):
		JList[l]=JList[l]/2.0

	Spin=0
	for l in n:
		Spin+=l
	Spin=2*abs(Spin)				

	return JList,Spin

def checkpresence(stuff,mylist):
	TF=0
	for n in mylist:
		if stuff == n[4]:
			TF=1
	return TF

def checkjeq(jeq,spin,mylist):
	TF=0
	for n in mylist:
		if jeq==n[2]:
			if int(spin)==n[0]:
				TF=1
	return TF

if __name__ == "__main__":
	Isomer=sys.argv[1]
	Terse=int(sys.argv[2])
	Predict=int(sys.argv[3])

## DON'T EDIT THIS FILE
## ALL DATA IS COLLECTED FROM SPINMATRICES.PY
	J,JName, SpinArray, JValue, JValueError = spinmatrices.getData(Isomer)
	
	if Predict==0:
		JValue= [0 if x=='0' else 1 for x in JName]
		JValueError= [0 if x=='0' else 1 for x in JName]

 
	JList=[0]*(max(max(J))+1)
	

#Check symmetry of spin matrix
	for n in range(0,len(J)):
		for m in range(0,len(J)): 
			if not (J[n][m]-J[m][n]==0):
				print 'Asymmetry', n,m

	#Generates all possible spin flip combinations
	AllSpecies=GenSpecies(J)
	#Generate linear equations and compute spins
	Species=[]
	
	for n in AllSpecies:
		Couplings, Spin=GenCoupling(n,J)
		
		Js=Couplings 
		JTot=0.0
		
		# Energy 
		EnergyDifference=0
		EnergyError=0
		Jeq=[0]*(getMax(J)+1)
		
		for m in range(0,(getMax(J)+1)):
			Jeq[m]+=Js[m]*JValue[m]
			JTot+=Jeq[m]
			EnergyDifference+=JValue[m]*Js[m]
			EnergyError+=JValueError[m]*JValueError[m]*Js[m]*Js[m]

		# Generate percentage contributions
		JContrib=[0]*(getMax(J)+1)
		for m in range(0,(getMax(J)+1)):
			if JTot==0:
				JContrib[m]=0.0
			else:
				JContrib[m]=float(Jeq[m]/JTot)

		#Find flipped spins
		positions=[i+1 for i,x in enumerate(n) if x < 0]
		Flipped=positions

		if len(Flipped)>(len(J)/2): #Fixed
			positions=list(set(range(1,len(J)+1))-set(Flipped)) #Fixed
			positions.sort()

		#Weed out dupes:
		if not checkpresence(positions,Species):
			if Terse==1:
				if not ( checkjeq(Jeq,Spin, Species)):	
					Species+=[[int(Spin)+1,EnergyDifference,Jeq,JContrib, positions,sqrt(EnergyError)]]
			elif Terse==0:
				Species+=[[int(Spin)+1,EnergyDifference,Jeq,JContrib, positions,sqrt(EnergyError)]]

#	print Species
	JNames=''
	for n in JName:
		if not n=='0':
			JNames+=str(n)+'\t'
	StringToWrite='#Spin \t E \t dE \t %s %s Spins \n' % (str(JNames),str(JNames))
	
	print StringToWrite
	for Specie in Species:
		
		StringToWrite=''
			
		StringToWrite='%s\t' % str(Specie[0])
		StringToWrite+='%5.2f\t' % float(Specie[1])
		StringToWrite+='%3.2f\t' % float(Specie[5])
		
		#J coupling equation
		Jeqstr=''
		
		for n in range(0,len(Specie[2])):
		
			if not JName[n]=='0':
				Jeqstr+="%d\t"%(Specie[2][n])
		StringToWrite+=Jeqstr

		#J contribution
		Jeqstr=''
		for n in range(0,len(Specie[3])):
			if not JName[n]=='0':
				Jeqstr+="%1.2f\t"%(Specie[3][n])

		StringToWrite+=Jeqstr
		StringToWrite+=str(Specie[4])
		print StringToWrite

